/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::ExampleClient
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigation/libraries/global_planning/AStar.h"
#include "Navigation/libraries/trajectory_control/TrajectoryFollowingController.h"
#include <Navigation/components/ExampleClient/ExampleClient.h>

// STD/STL
#include <chrono>
#include <thread>

exns::ExampleClient::ExampleClient()
{
    // pass
}

exns::ExampleClient::~ExampleClient()
{
    // pass
}

void
exns::ExampleClient::onInitComponent()
{
    // pass
}

void
exns::ExampleClient::onConnectComponent()
{
    task = new armarx::RunningTask<ExampleClient>(this, &ExampleClient::exampleNavigation);
    task->start();
}

void
exns::ExampleClient::onDisconnectComponent()
{
    const bool join = true;
    task->stop(join);
}

void
exns::ExampleClient::onExitComponent()
{
    // pass
}

std::string
exns::ExampleClient::getDefaultName() const
{
    return "ExampleClient";
}

void
exns::ExampleClient::exampleNavigation()
{
    using namespace std::chrono_literals;

    // Import relevant namespaces.
    using namespace armarx::nav;

    ARMARX_INFO << "Configuring navigator";

    // Create an example configuration valid for the following move* calls.
    configureNavigator(
        NavigationStackConfig()
            .general({} /*{.maxVel = VelocityLimits{.linear = 400 , .angular = 0.1}}*/)
            .globalPlanner(glob_plan::AStarParams())
            .trajectoryController(traj_ctrl::TrajectoryFollowingControllerParams()));

    // Example of registering a lambda as callback.
    navigator.onGoalReached([&]() { ARMARX_IMPORTANT << "Goal reached! (lambda-style)"; });

    // Example of registering a method as callback.
    navigator.onGoalReached([this] { goalReached(); });

    std::this_thread::sleep_for(1s);

    ARMARX_INFO << "Moving to goal pose";
    // Start moving to goal position using above config.
    Pose goal = Pose::Identity();
    goal.translation() << 2000, 1000, 0;

    navigator.moveTo(goal, NavigationFrame::Absolute);

    std::this_thread::sleep_for(15s);

    // Wait until goal is reached
    navigator.waitForGoalReached();
    ARMARX_INFO << "Goal 1 reached.";

    goal.translation() << -1500, 1000, 0;
    navigator.moveTo(goal, NavigationFrame::Absolute);

    std::this_thread::sleep_for(15s);

    // Wait until goal is reached
    navigator.waitForGoalReached();
    ARMARX_INFO << "Goal 2 reached.";

    goal.translation() << 4500, 4500, 0;
    navigator.moveTo(goal, NavigationFrame::Absolute);

    // Wait until goal is reached
    navigator.waitForGoalReached();
    ARMARX_INFO << "Goal 3 reached.";

    std::this_thread::sleep_for(15s);

    // TODO example with waypoints

    ARMARX_INFO << "Moving into certain direction.";
    // Start moving towards a direction
    navigator.moveTowards(Direction(100, 100, 0), NavigationFrame::Relative);

    std::this_thread::sleep_for(3s);

    ARMARX_INFO << "Pausing movement.";
    navigator.pause();
}

void
exns::ExampleClient::goalReached()
{
    ARMARX_IMPORTANT << "Goal reached! (method-style)";
}

armarx::PropertyDefinitionsPtr
exns::ExampleClient::createPropertyDefinitions()
{
    armarx::PropertyDefinitionsPtr def =
        new armarx::ComponentPropertyDefinitions(getConfigIdentifier());
    return def;
}
