#include "AStar.h"

#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "Navigation/libraries/global_planning/Point2Point.h"
#include <Navigation/libraries/algorithms/astar/AStarPlanner.h>
#include <Navigation/libraries/algorithms/smoothing/ChainApproximation.h>
#include <Navigation/libraries/algorithms/smoothing/CircularPathSmoothing.h>
#include <Navigation/libraries/conversions/eigen.h>
#include <Navigation/libraries/core/Trajectory.h>
#include <Navigation/libraries/global_planning/aron/AStarParams.aron.generated.h>

namespace armarx::nav::glob_plan
{

    // AStarParams

    Algorithms
    AStarParams::algorithm() const
    {
        return Algorithms::AStar;
    }

    aron::datanavigator::DictNavigatorPtr
    AStarParams::toAron() const
    {
        return arondto::AStarParams().toAron(); // TODO implement
    }

    AStarParams
    AStarParams::FromAron(const aron::datanavigator::DictNavigatorPtr& dict)
    {
        return AStarParams(); // TODO implement
    }

    // AStar

    AStar::AStar(const AStarParams& params, const core::Scene& ctx) :
        GlobalPlanner(ctx), params(params)
    {
    }

    auto
    robotCollisionModel()
    {
        VirtualRobot::CoinVisualizationFactory factory;
        const auto cylinder = factory.createCylinder(1500.F / 2, 2000.F);

        VirtualRobot::CollisionModelPtr collisionModel(new VirtualRobot::CollisionModel(cylinder));

        return collisionModel;
    }

    std::vector<Eigen::Vector2f>
    AStar::postProcessPath(const std::vector<Eigen::Vector2f>& path)
    {
        /// chain approximation
        algo::ChainApproximation approx(path,
                                        algo::ChainApproximation::Params{.distanceTh = 200.F});
        approx.approximate();
        const auto p = approx.approximatedChain();

        // visualizePath(p, "approximated", simox::Color::green());

        algo::CircularPathSmoothing smoothing;
        const auto points = smoothing.smooth(p);

        return points;
    }

    core::TrajectoryPtr
    AStar::plan(const core::Pose& goal)
    {
        ARMARX_TRACE;

        algo::astar::AStarPlanner planner(scene.robot, scene.staticScene->objects, 100.F);
        // planner.setRobotColModel("Platform-colmodel");

        // planner.setRobotColModel("Platform-colmodel");
        planner.setRobotCollisionModel(robotCollisionModel());

        const Eigen::Vector2f startPos = conv::to2D(scene.robot->getGlobalPosition());
        const Eigen::Vector2f goalPos = conv::to2D(goal.translation());

        const auto timeStart = IceUtil::Time::now();

        std::vector<Eigen::Vector2f> plan;
        try
        {
            plan = planner.plan(startPos, goalPos);
        }
        catch (...)
        {
            ARMARX_WARNING << "Could not plan collision-free path.";
            return nullptr;
        }
        const auto timeEnd = IceUtil::Time::now();

        ARMARX_IMPORTANT << "A* planning took " << (timeEnd - timeStart).toMilliSeconds() << " ms";
        ARMARX_IMPORTANT << "Path contains " << plan.size() << " points";

        if (plan.size() < 2) // failure
        {
            ARMARX_WARNING << "Could not plan collision-free path!";
            return nullptr;
        }

        ARMARX_TRACE;
        auto smoothPlan = postProcessPath(plan);
        ARMARX_IMPORTANT << "Smooth path contains " << smoothPlan.size() << " points";

        ARMARX_TRACE;
        // we need to strip the first and the last points from the plan as they encode the start and goal position
        smoothPlan.erase(smoothPlan.begin());
        smoothPlan.pop_back();

        ARMARX_TRACE;
        auto trajectory = core::Trajectory::FromPath(core::Pose(scene.robot->getGlobalPose()),
                                                     conv::to3D(smoothPlan),
                                                     goal,
                                                     params.linearVelocity);

        // TODO(fabian.reister): resampling of trajectory

        // FIXME param
        auto resampledTrajectory = trajectory.resample(500);
        ARMARX_IMPORTANT << "Resampled trajectory contains " << resampledTrajectory.points().size()
                         << " points";

        // TODO(fabian.reister): define optimization problem to optimize orientation
        // - start and end pose are fixed
        // - orientation diff between neighbor nodes should be minimized (smoothness)
        // - check if position could be optimized gently (this might require resampling during optimization)

        ARMARX_TRACE;
        return std::make_shared<core::Trajectory>(trajectory);
    }
} // namespace armarx::nav::glob_plan
