#include "Navigator.h"

#include <algorithm>

#include <Eigen/Geometry>

#include <SimoxUtility/math/convert/mat3f_to_rpy.h>
#include <VirtualRobot/Robot.h>

#include "ArmarXCore/core/logging/Logging.h"
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "Navigation/libraries/server/monitoring/GoalReachedMonitor.h"
#include <Navigation/libraries/core/types.h>
#include <Navigation/libraries/server/StackResult.h>

namespace armarx::nav::server
{
    void
    stopIfRunning(PeriodicTask<Navigator>::pointer_type& task)
    {
        if (task)
        {
            task->stop();
        }
    }

    Navigator::Navigator(const NavigatorConfig& config) : config{config}
    {
        ARMARX_CHECK_NOT_NULL(config.scene) << "The scene must be set!";
        ARMARX_CHECK_NOT_NULL(config.executor) << "The executor must be set!";
    }

    Navigator::~Navigator()
    {
        ARMARX_INFO << "Navigator destructor";
        stop();

        stopAllThreads();
    }

    void
    Navigator::moveTo(const std::vector<core::Pose>& waypoints,
                      core::NavigationFrame navigationFrame)
    {
        ARMARX_INFO << "Received moveTo() request.";

        std::vector<core::Pose> globalWaypoints;
        switch (navigationFrame)
        {
            case core::NavigationFrame::Absolute:
                globalWaypoints = waypoints;
                break;
            case core::NavigationFrame::Relative:
                globalWaypoints.reserve(waypoints.size());
                std::transform(std::begin(waypoints),
                               std::end(waypoints),
                               std::back_inserter(globalWaypoints),
                               [&](const core::Pose& p)
                               { return core::Pose(config.scene->robot->getGlobalPose()) * p; });
                break;
        }

        moveToAbsolute(globalWaypoints);
    }

    void
    Navigator::moveToAbsolute(const std::vector<core::Pose>& waypoints)
    {
        // if this navigator is in use, stop the movement ...
        pause();
        // ... and all threads
        stopAllThreads();

        // global planner
        ARMARX_INFO << "Planning global trajectory";
        ARMARX_CHECK_NOT_NULL(config.stack.globalPlanner);
        // TODO plan on multiple waypoints, ignoring waypoints for now
        // idea: compute multiple global trajectories, one for each segment between waypoints.
        res.globalTrajectory = config.stack.globalPlanner->plan(waypoints.back());

        if (not res.globalTrajectory)
        {
            ARMARX_WARNING << "No global trajectory. Cannot move.";
            return;
        }

        ARMARX_INFO << "Global planning completed. Will now start all required threads";

        // local planner
        if (config.stack.localPlanner)
        {
            ARMARX_INFO << "Planning local trajectory enabled";
            localPlannerTask =
                new PeriodicTask<Navigator>(this,
                                            &Navigator::updateLocalPlanner,
                                            config.general.tasks.localPlanningUpdatePeriod,
                                            false,
                                            "LocalPlannerTask");
            localPlannerTask->start();
        }

        // trajectory controller
        ARMARX_CHECK_NOT_NULL(config.stack.trajectoryControl);
        trajectoryControllerTask =
            new PeriodicTask<Navigator>(this,
                                        &Navigator::updateTrajectoryController,
                                        config.general.tasks.trajectoryControllerUpdatePeriod,
                                        false,
                                        "TrajectoryControllerTask");
        trajectoryControllerTask->start();

        // safety controller
        if (config.stack.safetyControl)
        {
            ARMARX_INFO << "Safety control enabled";
            safetyControllerTask =
                new PeriodicTask<Navigator>(this,
                                            &Navigator::updateSafetyController,
                                            config.general.tasks.safetyControllerUpdatePeriod,
                                            false,
                                            "SafetyControllerTask");
            safetyControllerTask->start();
        }

        // executor
        executorTask = new PeriodicTask<Navigator>(this,
                                                   &Navigator::updateExecutor,
                                                   config.general.tasks.executorUpdatePeriod,
                                                   false,
                                                   "ExecutorTask");
        executorTask->start();

        // introspector
        introspectorTask =
            new PeriodicTask<Navigator>(this,
                                        &Navigator::updateIntrospector,
                                        config.general.tasks.introspectorUpdatePeriod,
                                        false,
                                        "IntrospectorTask");
        introspectorTask->start();

        // monitoring
        goalReachedMonitor =
            GoalReachedMonitor(waypoints, *config.scene, GoalReachedMonitorConfig());
        monitorTask = new PeriodicTask<Navigator>(this,
                                                  &Navigator::updateMonitor,
                                                  config.general.tasks.monitorUpdatePeriod,
                                                  false,
                                                  "MonitorTask");
        monitorTask->start();

        // Could be required if pauseMovement() has been called in the past.
        resume();
    }

    void
    Navigator::moveTowards(const core::Direction& direction, core::NavigationFrame navigationFrame)
    {
    }

    void
    Navigator::moveTowardsAbsolute(const core::Direction& direction)
    {
    }

    void
    Navigator::updateLocalPlanner()
    {
        res.localTrajectory = config.stack.localPlanner->plan(res.globalTrajectory);
    }

    void
    Navigator::updateTrajectoryController()
    {
        res.controlVelocity = config.stack.trajectoryControl->control(res.trajectory());
    }

    void
    Navigator::updateSafetyController()
    {
        if (res.controlVelocity)
        {
            res.safeVelocity = config.stack.safetyControl->control(res.controlVelocity.value());
        }
    }

    void
    Navigator::updateExecutor()
    {
        if (isPaused() or isStopped())
        {
            // [[unlikely]]
            return;
        }

        if (not isStackResultValid())
        {
            // [[unlikely]]
            return;
        }

        const core::Rotation robot_R_world(config.scene->robot->getGlobalOrientation().inverse());

        ARMARX_INFO << deactivateSpam(100) << "Robot orientation "
                    << simox::math::mat3f_to_rpy(config.scene->robot->getGlobalOrientation()).z();

        core::Twist robotFrameVelocity;

        robotFrameVelocity.linear = robot_R_world * res.velocity().linear;
        // FIXME fix angular velocity
        robotFrameVelocity.angular = res.velocity().angular;

        ARMARX_INFO << deactivateSpam(100) << "velocity in robot frame "
                    << robotFrameVelocity.linear;

        config.executor->move(robotFrameVelocity);
    }

    void
    Navigator::updateIntrospector()
    {
        ARMARX_CHECK_NOT_NULL(config.introspector);

        if (res.isValid())
        {
            config.introspector->onStackResult(res);
        }
    }

    void
    Navigator::updateMonitor()
    {
        if (not isPaused() and goalReachedMonitor->goalReached())
        {
            // [[unlikely]]

            ARMARX_INFO << "Goal reached!";
            stop();

            // stop all threads, including this one
            stopAllThreads();
        }
    }

    void
    Navigator::stopAllThreads()
    {
        stopIfRunning(localPlannerTask);
        stopIfRunning(trajectoryControllerTask);
        stopIfRunning(safetyControllerTask);
        stopIfRunning(executorTask);
        stopIfRunning(introspectorTask);
        stopIfRunning(monitorTask);
    }

    bool
    Navigator::isStackResultValid() const noexcept
    {

        // global planner
        if (config.stack.globalPlanner != nullptr)
        {
            if (res.globalTrajectory == nullptr)
            {
                ARMARX_WARNING << deactivateSpam(1) << "Global trajectory not yet set.";
                return false;
            }
        }

        // local planner
        if (config.stack.localPlanner != nullptr)
        {
            if (res.localTrajectory == nullptr)
            {
                ARMARX_WARNING << deactivateSpam(1) << "Local trajectory not yet set.";
                return false;
            }
        }

        // trajectory controller
        if (config.stack.trajectoryControl != nullptr)
        {
            if (not res.controlVelocity.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Raw control velocity not yet set.";
                return false;
            }
        }

        // safety controller
        if (config.stack.safetyControl != nullptr)
        {
            if (not res.safeVelocity.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Safe velocity not yet set.";
                return false;
            }
        }

        // [[likely]]
        return true;
    }

    void
    Navigator::pause()
    {
        executorEnabled.store(false);

        ARMARX_CHECK_NOT_NULL(config.executor);
        config.executor->move(core::Twist::Zero());
    }

    void
    Navigator::resume()
    {
        executorEnabled.store(true);
    }

    void
    Navigator::stop()
    {
        pause();
        // TODO: Reset own state.
    }

    bool
    Navigator::isPaused() const noexcept
    {
        return executorEnabled.load();
    }

    bool
    Navigator::isStopped() const noexcept
    {
        // TODO: Implement.
        return false;
    }

    Navigator::Navigator(Navigator&& other) noexcept :
        config{std::move(other.config)}, executorEnabled{other.executorEnabled.load()}
    {
    }
} // namespace armarx::nav::server
